@@ -90,11 +90,16 @@ coro_init (void)
   coro_transfer (new_coro, create_coro);
 
 #if __GCC_HAVE_DWARF2_CFI_ASM && __amd64
-  asm (".cfi_undefined rip");
+  /*asm (".cfi_startproc");*/
+  /*asm (".cfi_undefined rip");*/
 #endif
 
   func ((void *)arg);
 
+#if __GCC_HAVE_DWARF2_CFI_ASM && __amd64
+  /*asm (".cfi_endproc");*/
+#endif
+
   /* the new coro returned. bad. just abort() for now */
   abort ();
 }
@@ -117,6 +122,13 @@ trampoline (int sig)
 
 # if CORO_ASM
 
+  #if __arm__ && \
+      (defined __ARM_ARCH_7__  || defined __ARM_ARCH_7A__ \
+    || defined __ARM_ARCH_7R__ || defined __ARM_ARCH_7M__ \
+    || __ARM_ARCH == 7)
+    #define CORO_ARM 1
+  #endif
+
   #if _WIN32 || __CYGWIN__
     #define CORO_WIN_TIB 1
   #endif
@@ -206,7 +218,7 @@ trampoline (int sig)
          "\tpopq %rcx\n"
          "\tjmpq *%rcx\n"
 
-       #elif __i386
+       #elif __i386__
 
          #define NUM_SAVED 4
          "\tpushl %ebp\n"
@@ -234,6 +246,80 @@ trampoline (int sig)
          "\tpopl %ecx\n"
          "\tjmpl *%ecx\n"
 
+       #elif CORO_ARM /* untested, what about thumb, neon, iwmmxt? */
+
+         #if __ARM_PCS_VFP
+           "\tvpush {d8-d15}\n"
+           #define NUM_SAVED (9 + 8 * 2)
+         #else
+           #define NUM_SAVED 9
+         #endif
+         "\tpush {r4-r11,lr}\n"
+         "\tstr sp, [r0]\n"
+         "\tldr sp, [r1]\n"
+         "\tpop {r4-r11,lr}\n"
+         #if __ARM_PCS_VFP
+           "\tvpop {d8-d15}\n"
+         #endif
+         "\tmov r15, lr\n"
+
+       #elif __mips__ && 0 /* untested, 32 bit only */
+
+        #define NUM_SAVED (12 + 8 * 2)
+         /* TODO: n64/o64, lw=>ld */
+
+         "\t.set    nomips16\n"
+         "\t.frame  $sp,112,$31\n"
+         #if __mips_soft_float
+           "\taddiu   $sp,$sp,-44\n"
+         #else
+           "\taddiu   $sp,$sp,-112\n"
+           "\ts.d     $f30,88($sp)\n"
+           "\ts.d     $f28,80($sp)\n"
+           "\ts.d     $f26,72($sp)\n"
+           "\ts.d     $f24,64($sp)\n"
+           "\ts.d     $f22,56($sp)\n"
+           "\ts.d     $f20,48($sp)\n"
+         #endif
+         "\tsw      $28,40($sp)\n"
+         "\tsw      $31,36($sp)\n"
+         "\tsw      $fp,32($sp)\n"
+         "\tsw      $23,28($sp)\n"
+         "\tsw      $22,24($sp)\n"
+         "\tsw      $21,20($sp)\n"
+         "\tsw      $20,16($sp)\n"
+         "\tsw      $19,12($sp)\n"
+         "\tsw      $18,8($sp)\n"
+         "\tsw      $17,4($sp)\n"
+         "\tsw      $16,0($sp)\n"
+         "\tsw      $sp,0($4)\n"
+         "\tlw      $sp,0($5)\n"
+         #if !__mips_soft_float
+           "\tl.d     $f30,88($sp)\n"
+           "\tl.d     $f28,80($sp)\n"
+           "\tl.d     $f26,72($sp)\n"
+           "\tl.d     $f24,64($sp)\n"
+           "\tl.d     $f22,56($sp)\n"
+           "\tl.d     $f20,48($sp)\n"
+         #endif
+         "\tlw      $28,40($sp)\n"
+         "\tlw      $31,36($sp)\n"
+         "\tlw      $fp,32($sp)\n"
+         "\tlw      $23,28($sp)\n"
+         "\tlw      $22,24($sp)\n"
+         "\tlw      $21,20($sp)\n"
+         "\tlw      $20,16($sp)\n"
+         "\tlw      $19,12($sp)\n"
+         "\tlw      $18,8($sp)\n"
+         "\tlw      $17,4($sp)\n"
+         "\tlw      $16,0($sp)\n"
+         "\tj       $31\n"
+         #if __mips_soft_float
+           "\taddiu   $sp,$sp,44\n"
+         #else
+           "\taddiu   $sp,$sp,112\n"
+         #endif
+
        #else
          #error unsupported architecture
        #endif
@@ -313,10 +399,10 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
 # elif CORO_LOSER
 
   coro_setjmp (ctx->env);
-  #if __CYGWIN__ && __i386
+  #if __CYGWIN__ && __i386__
     ctx->env[8]                        = (long)    coro_init;
     ctx->env[7]                        = (long)    ((char *)sptr + ssize)         - sizeof (long);
-  #elif __CYGWIN__ && __x86_64
+  #elif __CYGWIN__ && __x86_64__
     ctx->env[7]                        = (long)    coro_init;
     ctx->env[6]                        = (long)    ((char *)sptr + ssize)         - sizeof (long);
   #elif defined __MINGW32__
@@ -347,7 +433,7 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
   #elif defined (__GNU_LIBRARY__) && defined (__i386__)
     ctx->env[0].__jmpbuf[0].__pc       = (char *)  coro_init;
     ctx->env[0].__jmpbuf[0].__sp       = (void *)  ((char *)sptr + ssize)         - sizeof (long);
-  #elif defined (__GNU_LIBRARY__) && defined (__amd64__)
+  #elif defined (__GNU_LIBRARY__) && defined (__x86_64__)
     ctx->env[0].__jmpbuf[JB_PC]        = (long)    coro_init;
     ctx->env[0].__jmpbuf[0].__sp       = (void *)  ((char *)sptr + ssize)         - sizeof (long);
   #else
@@ -362,19 +448,34 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
 
 # elif CORO_ASM
 
-  ctx->sp = (void **)(ssize + (char *)sptr);
-  *--ctx->sp = (void *)abort; /* needed for alignment only */
-  *--ctx->sp = (void *)coro_init;
-
-  #if CORO_WIN_TIB
-  *--ctx->sp = 0;                    /* ExceptionList */
-  *--ctx->sp = (char *)sptr + ssize; /* StackBase */
-  *--ctx->sp = sptr;                 /* StackLimit */
+  #if __i386__ || __x86_64__
+    ctx->sp = (void **)(ssize + (char *)sptr);
+    *--ctx->sp = (void *)abort; /* needed for alignment only */
+    *--ctx->sp = (void *)coro_init;
+    #if CORO_WIN_TIB
+      *--ctx->sp = 0;                    /* ExceptionList */
+      *--ctx->sp = (char *)sptr + ssize; /* StackBase */
+      *--ctx->sp = sptr;                 /* StackLimit */
+    #endif
+  #elif CORO_ARM
+    /* return address stored in lr register, don't push anything */
+  #else
+    #error unsupported architecture
   #endif
 
   ctx->sp -= NUM_SAVED;
   memset (ctx->sp, 0, sizeof (*ctx->sp) * NUM_SAVED);
 
+  #if __i386__ || __x86_64__
+    /* done already */
+  #elif CORO_ARM
+    ctx->sp[0] = coro; /* r4 */
+    ctx->sp[1] = arg;  /* r5 */
+    ctx->sp[8] = (char *)coro_init; /* lr */
+  #else
+    #error unsupported architecture
+  #endif
+
 # elif CORO_UCONTEXT
 
   getcontext (&(ctx->uc));
@@ -406,15 +507,6 @@ struct coro_init_args
   coro_context *self, *main;
 };
 
-static pthread_t null_tid;
-
-/* I'd so love to cast pthread_mutex_unlock to void (*)(void *)... */
-static void
-mutex_unlock_wrapper (void *arg)
-{
-  pthread_mutex_unlock ((pthread_mutex_t *)arg);
-}
-
 static void *
 coro_init (void *args_)
 {
@@ -422,13 +514,8 @@ coro_init (void *args_)
   coro_func func = args->func;
   void *arg = args->arg;
 
-  pthread_mutex_lock (&coro_mutex);
-
-  /* we try to be good citizens and use deferred cancellation and cleanup handlers */
-  pthread_cleanup_push (mutex_unlock_wrapper, &coro_mutex);
-    coro_transfer (args->self, args->main);
-    func (arg);
-  pthread_cleanup_pop (1);
+  coro_transfer (args->self, args->main);
+  func (arg);
 
   return 0;
 }
@@ -436,11 +523,25 @@ coro_init (void *args_)
 void
 coro_transfer (coro_context *prev, coro_context *next)
 {
+  pthread_mutex_lock (&coro_mutex);
+
+  next->flags = 1;
   pthread_cond_signal (&next->cv);
-  pthread_cond_wait (&prev->cv, &coro_mutex);
-#if __FreeBSD__ /* freebsd is of course broken and needs manual testcancel calls... yay... */
-  pthread_testcancel ();
-#endif
+
+  prev->flags = 0;
+
+  while (!prev->flags)
+    pthread_cond_wait (&prev->cv, &coro_mutex);
+
+  if (prev->flags == 2)
+    {
+      pthread_mutex_unlock (&coro_mutex);
+      pthread_cond_destroy (&prev->cv);
+      pthread_detach (pthread_self ());
+      pthread_exit (0);
+    }
+
+  pthread_mutex_unlock (&coro_mutex);
 }
 
 void
@@ -453,9 +554,7 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
     {
       once = 1;
 
-      pthread_mutex_lock (&coro_mutex);
       pthread_cond_init (&nctx.cv, 0);
-      null_tid = pthread_self ();
     }
 
   pthread_cond_init (&ctx->cv, 0);
@@ -464,6 +563,7 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
     {
       pthread_attr_t attr;
       struct coro_init_args args;
+      pthread_t id;
 
       args.func = coro;
       args.arg  = arg;
@@ -481,26 +581,19 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, size_t ss
       pthread_attr_setstack (&attr, sptr, (size_t)ssize);
 #endif
       pthread_attr_setscope (&attr, PTHREAD_SCOPE_PROCESS);
-      pthread_create (&ctx->id, &attr, coro_init, &args);
+      pthread_create (&id, &attr, coro_init, &args);
 
       coro_transfer (args.main, args.self);
     }
-  else
-    ctx->id = null_tid;
 }
 
 void
 coro_destroy (coro_context *ctx)
 {
-  if (!pthread_equal (ctx->id, null_tid))
-    {
-      pthread_cancel (ctx->id);
-      pthread_mutex_unlock (&coro_mutex);
-      pthread_join (ctx->id, 0);
-      pthread_mutex_lock (&coro_mutex);
-    }
-
-  pthread_cond_destroy (&ctx->cv);
+  pthread_mutex_lock (&coro_mutex);
+  ctx->flags = 2;
+  pthread_cond_signal (&ctx->cv);
+  pthread_mutex_unlock (&coro_mutex);
 }
 
 /*****************************************************************************/
@@ -602,7 +695,7 @@ coro_destroy (coro_context *ctx)
 # undef CORO_GUARDPAGES
 #endif
 
-#if !__i386 && !__x86_64 && !__powerpc && !__m68k && !__alpha && !__mips && !__sparc64
+#if !__i386__ && !__x86_64__ && !__powerpc__ && !__arm__ && !__aarch64__ && !__m68k__ && !__alpha__ && !__mips__ && !__sparc64__
 # undef CORO_GUARDPAGES
 #endif
 
@@ -674,7 +767,7 @@ coro_stack_alloc (struct coro_stack *stack, unsigned int size)
   #endif
 
   #if CORO_USE_VALGRIND
-    stack->valgrind_id = VALGRIND_STACK_REGISTER ((char *)base, (char *)ssze - CORO_GUARDPAGES * PAGESIZE);
+    stack->valgrind_id = VALGRIND_STACK_REGISTER ((char *)base, ((char *)base) + ssze - CORO_GUARDPAGES * PAGESIZE);
   #endif
 
   stack->sptr = base;
